/*
 * (c) Copyright 2020 CORSIKA Project, corsika-project@lists.kit.edu
 *
 * This software is distributed under the terms of the GNU General Public
 * Licence version 3 (GPL Version 3). See file LICENSE for a full version of
 * the license.
 */

#pragma once

#include <corsika/framework/process/BaseProcess.hpp>
#include <corsika/framework/core/PhysicalUnits.hpp>

#include <corsika/detail/framework/process/InteractionProcess.hpp> // for extra traits, method/interface checking

namespace corsika {

  /**
     @ingroup Processes     
     @{

     Process describing the interaction of particles

     The structural base type for any interaction process in a
     ProcessSequence. 

     Create a new InteractionProcess, e.g. for XYModel, via 
     \code
     class XYModel : public InteractionProcess<XYModel> {};
     \endcode

     and provide the two necessary interface methods
     \code
     template <typename TSecondaryView> 
     void doInteraction(TSecondaryView&);

     template <typename TParticle>
     GrammageType getInteractionLength(TParticle const&)
     \endcode
     
     Where, of course, SecondaryView and Particle are the valid
     classes to access particles on the Stack. Those two methods do
     not need to be templated, they could use the types
     e.g. corsika::setup::Stack::particle_type -- but by the cost of
     loosing all flexibility otherwise provided. 

     (For your potential interest, InteractionProcess is based on the
     CRTP C++ design pattern)
   */

  template <typename TDerived>
  class InteractionProcess : public BaseProcess<TDerived> {

  public:
    using BaseProcess<TDerived>::ref;

    template <typename TParticle>
    InverseGrammageType getInverseInteractionLength(TParticle const& particle) {
      
      // interface checking on TProcess1
      static_assert(has_method_getInteractionLength_v<TDerived, GrammageType, TParticle const&>,
		    "TDerived has no method with correct signature \"GrammageType "
		    "getInteractionLength(TParticle const&)\" required for "
		    "InteractionProcess<TDerived>. ");
      
      return 1. / ref().getInteractionLength(particle);
    }
  };

  /** @} */
  
} // namespace corsika
